import pytest

from pytsql.tsql import _split

from .helpers import assert_strings_equal_disregarding_whitespace


@pytest.mark.parametrize(
    "seed",
    [
        """
        IF OBJECT_ID ( 'test' , 'V' ) IS NOT NULL
        DROP VIEW test""",
        """
        IF DATENAME ( weekday , GETDATE ( ) ) IN ( 'Saturday' , 'Sunday' )
            SELECT 'Weekend' ;
        ELSE
            SELECT 'Weekday' ;""",  # noqa: W291
        """
        IF OBJECT_ID ( 'test' , 'V' ) IS NOT NULL
            IF DATENAME ( weekday , GETDATE ( ) ) IN ( N'Saturday' , N'Sunday' )
                SELECT 'Weekend' ;
            ELSE
                SELECT 'Weekday' ;""",  # noqa: W291
        """
        WHILE OBJECT_ID ( 'test' , 'V' ) IS NOT NULL
        DROP VIEW test""",  # noqa: W291
        """
        WHILE 8 % 2 = 0
            IF 4 % 2 = 0
                SELECT ( 'HI' )
            ELSE
                SELECT ( 'HO' )
        """,  # noqa: W291
    ],
)
def test_control_flow(seed):
    splits = _split(seed)
    assert len(splits) == 1
    assert_strings_equal_disregarding_whitespace(splits[0], seed)



from pytsql.tsql import _split


def test_remove_comment_block():
    seed = """
USE master;
/*
Here is a comment we should not see
*/
select * from [master].[dbo].[table];
    """
    expected = """USE master ;
select * from [master] . [dbo] . [table] ;"""
    assert _process(seed) == expected


def test_remove_comment_line():
    seed = """
-- USE master;
select * from [master].[dbo].[table];
"""
    expected = """select * from [master] . [dbo] . [table] ;"""
    assert _process(seed) == expected


def test_remove_trailing_comment():
    seed = """
USE master;
select * -- select everything!
from [master].[dbo].[table];
"""
    expected = """USE master ;
select * from [master] . [dbo] . [table] ;"""
    assert _process(seed) == expected


def test_remove_single_line_block():
    seed = """
USE master;
/* This comment has great intentions. */
select *
from [master].[dbo].[table];"""
    expected = """USE master ;
select * from [master] . [dbo] . [table] ;"""
    assert _process(seed) == expected


def test_single_line_wrapping_blocks():
    seed = """USE master;
/* Start comment */ select * /* End comment */
from [master].[dbo].[table];"""
    expected = """USE master ;
select * from [master] . [dbo] . [table] ;"""
    assert _process(seed) == expected


def test_comments_in_single_quotes():
    seed = """USE master;
select 'foo -- bar /* text */'
from [master].[dbo].[table];"""
    expected = """USE master ;
select 'foo -- bar /* text */' from [master] . [dbo] . [table] ;"""
    assert _process(seed) == expected


def test_comments_in_double_quotes():
    seed = """USE master;
select "foo -- bar /* text */"
from [master].[dbo].[table];"""
    expected = """USE master ;
select "foo -- bar /* text */" from [master] . [dbo] . [table] ;"""
    assert _process(seed) == expected


def test_comments_in_string_literal_with_escaped_single_quote():
    seed = r"""USE master;
select 'foo -- '' bar /* text */'
from [master].[dbo].[table];"""
    expected = r"""USE master ;
select 'foo -- '' bar /* text */' from [master] . [dbo] . [table] ;"""
    assert _process(seed) == expected


def _process(seed):
    return "\n".join(_split(seed))



import inspect
import re
from typing import List
from unittest import mock

import pytest
from antlr4 import InputStream, ParserRuleContext
from antlr4.Token import CommonToken
from antlr4.tree.Tree import TerminalNodeImpl

from pytsql import tsql
from pytsql.grammar import sa_tsql
from pytsql.tsql import _split


@pytest.fixture
def seed():
    return """
DECLARE @timestamp_started  DATETIME  = CURRENT_TIMESTAMP;
-- step 0: deduce associated nr's
DROP TABLE IF EXISTS #temporary;
CREATE TABLE #temporary (
    nr      VARCHAR(17)   NOT NULL
  , short_nr         SMALLINT NOT NULL
  , min_nr_date     DATE     NOT NULL
  , iter        TINYINT  NOT NULL -- iterator
  , PRIMARY KEY (nr, min_date, iter)
);
PRINT(CAST(GETDATE() AS VARCHAR) + ': INSERT INTO #temporary')
INSERT INTO #temporary WITH(TABLOCK)
SELECT CONCAT('0', nr) nr
     , short_nr
     , MIN(nr_date) min_nr_date
     , ROW_NUMBER() OVER(PARTITION BY nr, MIN(nr_date) ORDER BY short_nr DESC) -- very few instances where this matters
FROM otherdb.dbo.othertable WITH(NOLOCK)
WHERE kind = 1
GROUP BY nr, short_nr;"""


def get_rule_labels(context_cls: ParserRuleContext) -> List[str]:
    init_func = context_cls.__init__

    # Detect any context/token labels from the init function's assignments
    lines, _ = inspect.getsourcelines(init_func)
    labels = []
    for line in lines:
        m = re.match(r"self\.(\w+)\s*=\s*None", line.strip())
        if m:
            labels.append(m.group(1))

    return labels


def compare_common_token(py_token: CommonToken, cpp_token: CommonToken) -> None:
    assert type(py_token) == type(cpp_token)
    if py_token is None:
        return
    assert py_token.type == cpp_token.type
    assert py_token.channel == cpp_token.channel
    assert py_token.start == cpp_token.start
    assert py_token.stop == cpp_token.stop
    assert py_token.tokenIndex == cpp_token.tokenIndex
    assert py_token.line == cpp_token.line
    assert py_token.column == cpp_token.column
    assert py_token.text == cpp_token.text
    assert py_token.getInputStream() is cpp_token.getInputStream()


def compare_terminal_node(
    py_terminal_node: TerminalNodeImpl, cpp_terminal_node: TerminalNodeImpl
) -> None:
    assert type(py_terminal_node) is type(cpp_terminal_node)
    assert type(py_terminal_node.symbol) is type(cpp_terminal_node.symbol)  # noqa: E721

    compare_common_token(py_terminal_node.symbol, cpp_terminal_node.symbol)


def compare_context(
    py_context: ParserRuleContext, cpp_context: ParserRuleContext
) -> None:
    assert type(py_context) == type(cpp_context)
    assert len(py_context.children) == len(cpp_context.children)

    # Validate children
    for i in range(len(py_context.children)):
        if isinstance(py_context.children[i], TerminalNodeImpl):
            compare_terminal_node(py_context.children[i], cpp_context.children[i])
        elif isinstance(py_context.children[i], ParserRuleContext):
            compare_context(py_context.children[i], cpp_context.children[i])
        else:
            raise TypeError(f"Unexpected node type: {py_context.children[i]}")
        assert py_context.children[i].parentCtx is py_context
        assert cpp_context.children[i].parentCtx is cpp_context

    # Validate labels
    for label in get_rule_labels(py_context):
        py_label = getattr(py_context, label)
        cpp_label = getattr(cpp_context, label)
        assert type(py_label) == type(cpp_label)
        if isinstance(py_label, CommonToken):
            compare_common_token(py_label, cpp_label)
        elif isinstance(py_label, ParserRuleContext):
            compare_context(py_label, cpp_label)


def test_compare_py_and_cpp_parse_trees(seed):
    """Compares the trees produced by Python and C++ version of the parser

    Based on validation code included in `speedy_antlr_tool`
    https://github.com/amykyta3/speedy-antlr-tool/blob/master/speedy_antlr_tool/validate.py

    Some assertions have been removed to account for (closed but still present) issue
    https://github.com/amykyta3/speedy-antlr-tool/issues/3
    """

    stream = InputStream(data=seed)

    py_context = sa_tsql._py_parse(stream, "tsql_file")
    cpp_context = sa_tsql._cpp_parse(stream, "tsql_file")

    assert py_context.parentCtx is None
    assert cpp_context.parentCtx is None

    compare_context(py_context, cpp_context)


@mock.patch("pytsql.grammar.sa_tsql._cpp_parse", wraps=sa_tsql._cpp_parse)
@mock.patch("pytsql.grammar.sa_tsql._py_parse", wraps=sa_tsql._py_parse)
def test_compare_py_and_cpp__split(py_parse_mock, cpp_parse_mock, seed):
    sa_tsql.USE_CPP_IMPLEMENTATION = False
    py_result = _split(seed)
    py_parse_mock.assert_called_once()

    sa_tsql.USE_CPP_IMPLEMENTATION = True
    cpp_result = _split(seed)
    cpp_parse_mock.assert_called_once()

    assert py_result == cpp_result


def test__py_parse__warns():
    tsql.USE_CPP_IMPLEMENTATION = False
    with pytest.warns(match=r"Can not find C\+\+ version of the parser"):
        _split("")



import logging

from pytsql import executes


def test_evaluations(engine, caplog):
    caplog.set_level(logging.INFO)

    seed = """
    PRINT 'basic text'
    PRINT(10 + 10)
    PRINT('two plus two equals ' + CAST(4 AS VARCHAR))
    """

    executes(seed, engine)

    assert "basic text" in caplog.text
    assert "20" in caplog.text
    assert "two plus two equals 4" in caplog.text
    assert caplog.text.count("Captured PRINT") == 3


def test_evaluation_in_cf(engine, caplog):
    caplog.set_level(logging.INFO)
    seed = """
    IF 1 = 1
        BEGIN
            PRINT 'text1'
            PRINT 'text2'
            PRINT 'text3'
        END
    """

    executes(seed, engine)

    assert "text1" in caplog.text
    assert "text2" in caplog.text
    assert "text3" in caplog.text
    assert caplog.text.count("Captured PRINT") == 3


def test_evaluation_in_context(engine, caplog):
    caplog.set_level(logging.INFO)

    seed = """
    DECLARE @A INT = 0
    WHILE @A < 10
        BEGIN
            PRINT @A
            SET @A = @A + 1
        END
    """

    executes(seed, engine)

    for i in range(10):
        assert str(i) in caplog.text
    assert caplog.text.count("Captured PRINT") == 10


def test_print_unicode(engine, caplog):
    caplog.set_level(logging.INFO)

    unicode_text = "тест"
    seed = f"PRINT N'{unicode_text}'"

    executes(seed, engine)

    assert unicode_text in caplog.text


def test_print_truncation(engine, caplog):
    caplog.set_level(logging.INFO)

    print_text = "a" * 5000
    seed = f"PRINT '{print_text}'"

    executes(seed, engine)

    assert ("a" * 2000) in caplog.text
    assert ("a" * 2001) not in caplog.text


def test_multiple_prints(engine):
    for _ in range(50):
        executes("SELECT 12", engine)



from pytsql.tsql import _split

from .helpers import assert_strings_equal_disregarding_whitespace


def test_declaration_in_control_flow():
    seed = """
    IF 1 = 1
        DECLARE @A INT = 5
    SELECT * FROM x
    """
    splits = _split(seed)
    assert len(splits) == 2

    assert_strings_equal_disregarding_whitespace(
        splits[0], "IF 1 = 1 DECLARE @A INT = 5"
    )
    assert_strings_equal_disregarding_whitespace(
        splits[1],
        """
        DECLARE @A INT = 5
        SELECT * FROM x
        """,
    )


def test_dont_append_dynamics_on_create_schema():
    seed = """
        DECLARE @A INT = 5
        CREATE SCHEMA s
        SELECT * FROM x
        CREATE VIEW y AS SELECT 1
        """
    splits = _split(seed)
    assert len(splits) == 4

    assert splits[0] == "DECLARE @A INT = 5"
    assert splits[1] == "CREATE SCHEMA s"
    assert_strings_equal_disregarding_whitespace(
        splits[2],
        """
        DECLARE @A INT = 5
        SELECT * FROM x
        """,
    )
    assert splits[3] == "CREATE VIEW y AS SELECT 1"


def test_dynamics_no_isolation():
    seed = """
    DECLARE @A INT = 5
    SELECT @A
    GO
    SELECT @A
    """
    splits = _split(seed, isolate_top_level_statements=False)
    assert len(splits) == 2
    assert_strings_equal_disregarding_whitespace(
        splits[0],
        """
        DECLARE @A INT = 5
        SELECT @A
        """,
    )
    assert_strings_equal_disregarding_whitespace(
        splits[1],
        """
        DECLARE @A INT = 5
        SELECT @A
        """,
    )



import logging
import re
import warnings
from pathlib import Path
from re import Match
from typing import Any, Dict, List, Optional, Union

import antlr4.tree.Tree
import sqlalchemy
from antlr4 import InputStream, Token
from sqlalchemy.engine import Connection

from pytsql.grammar import USE_CPP_IMPLEMENTATION, SA_ErrorListener, parse, tsqlParser

_REPLACE_START = "<replace>"
_REPLACE_END = "</replace>"
_EOF = "<EOF>"
_GO = "GO"
_PRINTS_TABLE = "#pytsql_prints"


def _code(path: Union[str, Path], encoding: str) -> str:
    with Path(path).open(encoding=encoding) as fh:
        return "\n".join(fh.readlines())


def _process_replacement(line: str, parameters: Dict[str, Any]) -> str:
    """Appropriately replace a single <replace> statement."""
    new_line = line.format(**parameters)
    if None in parameters.values():
        raise ValueError("Found a parameter with no specified replacement value.")
    if "{" and "}" in new_line:
        raise ValueError(
            "There was a parameter that was not replaced. Double check the parameters dictionary."
        )
    return new_line


def _parametrize(
    source: str,
    parameters: Dict[str, Any],
    start: str = _REPLACE_START,
    end: str = _REPLACE_END,
) -> str:
    """Replace all {start} and {end} statements, i.e. parametrizes the SQL script.

    Parameters
    ----------
        source:     string containing the source SQL code
        parameters: dictionary containing variables and their replacements
    Returns
    -------
        the input source code, separated by newlines, with parameter replacements
    """

    def parametrization_replacer(match: Match) -> str:
        return _process_replacement(match.group(1), parameters)

    # The pattern matches all parametrization patterns, including those within a string literal.
    pattern = re.compile(
        rf"/\* {re.escape(start)} \*/.*?/\* {re.escape(end)}(.*?) \*/",
        re.DOTALL | re.MULTILINE,
    )

    parametrized = re.sub(pattern, parametrization_replacer, source)

    non_empty_stripped_lines = [
        x.strip() for x in parametrized.split("\n") if x.strip() != ""
    ]

    return "\n".join(non_empty_stripped_lines)


class _TSQLVisitor(antlr4.ParseTreeVisitor):
    def __init__(self):
        # This attribute holds declarations of sql variables, also
        # referred to as `dynamic sql`.
        # In sql ides, these variables remain part of a session and
        # can therefore be used in the entirety of a script.
        # When executing code via sqlalchemy, on the other hand, this
        # session information is lost after the execution of a batch.
        # We therefore need to manually prepend it to all following
        # batches.
        self.dynamics: List[str] = []

    def visit(
        self, tree: tsqlParser.Sql_clauseContext, prepend_dynamics: bool = True
    ) -> str:
        dynamics = self.dynamics[:]

        chunks = tree.accept(self)

        # CREATE SCHEMA/VIEW must be the only statement in a batch
        is_create_schema_or_view = tree.ddl_clause() is not None and (
            tree.ddl_clause().create_schema() is not None
            or tree.ddl_clause().create_view() is not None
        )
        if not prepend_dynamics or is_create_schema_or_view:
            return " ".join(chunks)

        return " ".join(dynamics + chunks)

    def visitChildren(self, node: antlr4.ParserRuleContext) -> List[str]:
        if isinstance(node, tsqlParser.Print_statementContext):
            # Print statements are replaced by inserts into a temporary table so that they can be evaluated
            # at the right time and fetched afterwards.
            result = (
                f"INSERT INTO {_PRINTS_TABLE} VALUES ( LEFT (".split()
                + super().visitChildren(node.expression())
                + ", 2000 ) ) ;".split()
            )
        else:
            result = super().visitChildren(node)

        if isinstance(node, tsqlParser.Declare_statementContext):
            self.dynamics.extend(result)

        return result

    def visitTerminal(self, node: antlr4.TerminalNode) -> List[str]:
        return [str(node)]

    def defaultResult(self) -> List[str]:
        return []

    def aggregateResult(
        self, aggregate: List[str], next_result: List[str]
    ) -> List[str]:
        return aggregate + next_result


class _RaisingErrorListener(SA_ErrorListener):
    def syntaxError(
        self,
        input_stream: InputStream,
        offendingSymbol: Optional[Token],
        char_index: int,
        line: int,
        column: int,
        msg: str,
    ):
        error_message = f"Line {line}:{column} {msg}."
        if offendingSymbol is not None:
            error_message += f" Problem near token '{offendingSymbol.text}'."
        logging.error(error_message)
        raise ValueError(f"Error parsing SQL script: {error_message}")


def _split(code: str, isolate_top_level_statements: bool = True) -> List[str]:
    if not USE_CPP_IMPLEMENTATION:
        warnings.warn(
            "Can not find C++ version of the parser, Python version will be used instead."
            " Something is likely wrong with your installation of the package,"
            " and is preventing the use of the faster C++ parser."
        )

    logging.info("Started SQL script parsing")

    # The default error listener only prints to the console without raising exceptions.
    error_listener = _RaisingErrorListener()

    # Using code created by `speedy-antlr-tool` to parse the input.
    tree = parse(InputStream(data=code), "tsql_file", error_listener)
    visitor = _TSQLVisitor()

    # Our current definition of a 'batch' in isolation mode is a single top-level SQL clause.
    # Note that this differs from the grammar definition of a batch, which is
    # a group of clauses between GO statements. The latter matches the definition of batches
    # in non-isolation mode.
    batches = []
    for batch in tree.batch():
        clauses = []
        first_clause_in_batch = True
        for sql_clause in batch.sql_clauses().sql_clause():
            prepend_dynamics = first_clause_in_batch or isolate_top_level_statements
            clause = visitor.visit(sql_clause, prepend_dynamics=prepend_dynamics)
            if clause != "":
                clauses.append(clause)
                first_clause_in_batch = False
        if isolate_top_level_statements:
            batches.extend(clauses)
        else:
            batches.append("\n".join(clauses))

    logging.info("SQL script parsed successfully.")

    return batches


def _fetch_and_clear_prints(conn: Connection):
    prints = conn.execute(f"SELECT * FROM {_PRINTS_TABLE};")
    for row in prints.all():
        logging.info(f"Captured PRINT statement: {row[0]}")
    conn.execute(f"DELETE FROM {_PRINTS_TABLE};")


def executes(
    code: str,
    engine: sqlalchemy.engine.Engine,
    parameters: Optional[Dict[str, Any]] = None,
    isolate_top_level_statements=True,
) -> None:
    """Execute a given sql string through a sqlalchemy.engine.Engine connection.

    Please note either no parameters should be used or all parameters marked in a sql
    script should be given.

    Args
    ----
    code T-SQL string to be executed
    engine (sqlalchemy.engine.Engine): established mssql connection
    parameters An optional dictionary of parameters to substituted in the sql script
    isolate_top_level_statements: whether to execute statements one by one or in whole batches

    Returns
    -------
    None

    """
    parametrized_code = _parametrize(code, parameters) if parameters else code
    with engine.connect().execution_options(isolation_level="AUTOCOMMIT") as conn:
        # Since the prints table is a temporary one, it will be local to the connection and it will be dropped once the
        # connection is closed. Caveat: sqlalchemy engines can pool connections, so we still have to drop it preemtively.
        conn.execute(f"DROP TABLE IF EXISTS {_PRINTS_TABLE}")
        conn.execute(f"CREATE TABLE {_PRINTS_TABLE} (p NVARCHAR(4000))")
        for batch in _split(parametrized_code, isolate_top_level_statements):
            conn.execute(batch)
            _fetch_and_clear_prints(conn)


def execute(
    path: Union[str, Path],
    engine: sqlalchemy.engine.Engine,
    parameters: Optional[Dict[str, Any]] = None,
    isolate_top_level_statements=True,
    encoding: str = "utf-8",
) -> None:
    """Execute a given sql script through a sqlalchemy.engine.Engine connection.

    Args
    ----
    path (Path or str): Path to the sql file to be executed
    engine (sqlalchemy.engine.Engine): established mssql connection
    encoding: file encoding of the sql script (default: utf-8)
    isolate_top_level_statements: whether to execute statements one by one or in whole batches

    Returns
    -------
    None

    """
    executes(_code(path, encoding), engine, parameters, isolate_top_level_statements)



import logging

from pytsql import executes


def test_rowcount(engine, caplog):
    caplog.set_level(logging.INFO)

    seed = """
    USE [tempdb]
    GO
    DROP TABLE IF EXISTS [test_table]
    CREATE TABLE [test_table] (
        col VARCHAR(3)
    )
    GO
    INSERT INTO [test_table] (col)
    VALUES ('A'), ('AB'), ('ABC')
    PRINT('Affected ' + CAST(@@ROWCOUNT AS VARCHAR) + ' rows')
    """

    executes(seed, engine, isolate_top_level_statements=False)

    assert "Affected 3 rows" in caplog.text


def test_semi_persistent_set(engine, caplog):
    caplog.set_level(logging.INFO)

    seed = """
    DECLARE @A INT = 12
    DECLARE @B INT = 34
    SET @A = 56
    SET @B = 78
    PRINT(@A)
    GO
    PRINT(@B)
    """

    executes(seed, engine, isolate_top_level_statements=False)

    assert "56" in caplog.text
    assert "34" in caplog.text
    assert "12" not in caplog.text
    assert "78" not in caplog.text



def assert_strings_equal_disregarding_whitespace(string1: str, string2: str) -> None:
    string1_stripped = " ".join(string1.split())
    string2_stripped = " ".join(string2.split())
    assert string1_stripped == string2_stripped



import pytest

from pytsql.tsql import _split

from .helpers import assert_strings_equal_disregarding_whitespace


def test_ignore_go_inside_quoted_text():
    seed = """DECLARE @td            VARCHAR(20)   = CONVERT(VARCHAR, GETDATE(), 112);
DECLARE @dynamic       VARCHAR(8000) = '
CREATE DATABASE ' + @database_name +'
ON
( NAME = ' + @database_name +',
FILENAME = ''D:\\Microsoft SQL Server\\MSSQL13.MSSQLSERVER\\MSSQL\\DATA\' + @td +'_' + @database_name + '.mdf'',
SIZE = 1GB,
MAXSIZE = UNLIMITED,
FILEGROWTH = 100MB )
LOG ON
( NAME = ' + @database_name +'_log,
FILENAME = ''D:\\Microsoft SQL Server\\MSSQL13.MSSQLSERVER\\MSSQL\\DATA\' + @td +'_' + @database_name + '_log.ldf'',
SIZE = 10MB,
MAXSIZE = UNLIMITED,
FILEGROWTH = 10MB );'
;
SELECT(@dynamic)
EXEC(@dynamic)
"""
    splits = _split(seed)
    td_declare = (
        "DECLARE @td VARCHAR ( 20 ) = CONVERT ( VARCHAR , GETDATE ( ) , 112 ) ;"
    )
    assert_strings_equal_disregarding_whitespace(splits[0], td_declare)
    dynamic_declare = """
        DECLARE @dynamic       VARCHAR ( 8000 )  = '
        CREATE DATABASE ' + @database_name + '
        ON
        ( NAME = ' + @database_name + ',
        FILENAME = ''D:\\Microsoft SQL Server\\MSSQL13.MSSQLSERVER\\MSSQL\\DATA\' + @td + '_' + @database_name + '.mdf'',
        SIZE = 1GB,
        MAXSIZE = UNLIMITED,
        FILEGROWTH = 100MB )
        LOG ON
        ( NAME = ' + @database_name + '_log,
        FILENAME = ''D:\\Microsoft SQL Server\\MSSQL13.MSSQLSERVER\\MSSQL\\DATA\' + @td + '_' + @database_name + '_log.ldf'',
        SIZE = 10MB,
        MAXSIZE = UNLIMITED,
        FILEGROWTH = 10MB );'
        ;"""
    assert_strings_equal_disregarding_whitespace(
        splits[1], " ".join([td_declare, dynamic_declare])
    )
    assert_strings_equal_disregarding_whitespace(
        splits[2], " ".join([td_declare, dynamic_declare, "SELECT ( @dynamic )"])
    )
    assert_strings_equal_disregarding_whitespace(
        splits[3], " ".join([td_declare, dynamic_declare, "EXEC ( @dynamic )"])
    )


def test_select_concat():
    seed = """
DECLARE @timestamp_started  DATETIME  = CURRENT_TIMESTAMP;
-- step 0: deduce associated nr's
DROP TABLE IF EXISTS #temporary;
CREATE TABLE #temporary (
    nr      VARCHAR(17)   NOT NULL
  , short_nr         SMALLINT NOT NULL
  , min_nr_date     DATE     NOT NULL
  , iter        TINYINT  NOT NULL -- iterator
  , PRIMARY KEY (nr, min_date, iter)
);
PRINT(CAST(GETDATE() AS VARCHAR) + ': INSERT INTO #temporary')
INSERT INTO #temporary WITH(TABLOCK)
SELECT CONCAT('0', nr) nr
     , short_nr
     , MIN(nr_date) min_nr_date
     , ROW_NUMBER() OVER(PARTITION BY nr, MIN(nr_date) ORDER BY short_nr DESC) -- very few instances where this matters
FROM otherdb.dbo.othertable WITH(NOLOCK)
WHERE kind = 1
GROUP BY nr, short_nr;"""
    assert len(_split(seed)) == 5


def test_insert_select_from():
    seed = """
INSERT INTO #temporary WITH(TABLOCK)
SELECT o.nr
     , o.nr19
     , COALESCE(i1.short_nr, i2.short_nr, i_help1.short_nr, i_help2.short_nr) short_nr
     , COALESCE(i1.min_valid_from, i2.min_valid_from, i_help1.date_change, i_help2.date_change) min_valid_from_date
FROM (
  SELECT CONCAT('0', Nr) nr
       , nr19
       , MIN(VALID_FROM) min_valid_from
       , ROW_NUMBER() OVER(PARTITION BY Nr, MIN(VALID_FROM) ORDER BY COUNT(distinct VALID_FROM) DESC, nr19) iter
  FROM otherdb.dbo.history WITH(NOLOCK)
  CROSS APPLY ( VALUES (
    CASE LEFT(NR_PROZEDUR, 1) WHEN '5' THEN LEFT(NR_PROZEDUR, 6) + SUBSTRING(NR_PROZEDUR, 9, 13)
    ELSE LEFT(NR_PROZEDUR, 19) END
  )) nr(nr19)
  GROUP BY Nr, nr19
) o
LEFT JOIN (
  SELECT CONCAT('0', Nr ) nr
       , nr19
       , MIN(ISNULL(DATUM_LETZTER_FACHL_AEND, VALID_FROM)) date_change
       , ROW_NUMBER() OVER(PARTITION BY Nr, MIN(VALID_FROM) ORDER BY COUNT(distinct VALID_FROM) DESC, nr19) iter
  FROM otherdb.dbo.history WITH(NOLOCK)
  CROSS APPLY ( VALUES (
    CASE LEFT(NR_PROZEDUR, 1) WHEN '5' THEN LEFT(NR_PROZEDUR, 6) + SUBSTRING(NR_PROZEDUR, 9, 13)
    ELSE LEFT(NR_PROZEDUR, 19) END
  )) nr(nr19)
  WHERE drg <> ''
  GROUP BY Nr, nr19
) o_help
  ON  o.nr = o_help.nr
  AND o.nr19    = o_help.nr19
LEFT JOIN #temporary2Valid_From i1
  ON  o.nr = i1.nr
  AND o.min_valid_from   = i1.min_valid_from
  AND o.iter     = i1.iter
LEFT JOIN #temporary2Valid_From i2
  ON  o.nr = i2.nr
  AND o.min_valid_from   = i2.min_valid_from
  AND 1          = i2.iter
LEFT JOIN #temporary2Valid_From_help i_help1
  ON  o_help.nr    = i_help1.nr
  AND o_help.date_change = i_help1.date_change
  AND o_help.iter        = i_help1.iter
LEFT JOIN #temporary2Valid_From_help i_help2
  ON  o_help.nr    = i_help2.nr
  AND o_help.date_change = i_help2.date_change
  AND 1                  = i_help2.iter
WHERE COALESCE(i1.short_nr, i2.short_nr, i_help1.short_nr, i_help2.short_nr) IS NOT NULL;"""
    assert len(_split(seed)) == 1


def test_create_index_include_short():
    seed = """USE new_db
GO

/*
Create indices
*/
CREATE NONCLUSTERED INDEX output_index_1
ON [sim].[output1] ([chain])
INCLUDE ([id],[year],[iter])
GO

CREATE NONCLUSTERED INDEX output_index_2
ON [dbo].[output2] ([chain])
INCLUDE ([id],[year],[iter],[iter2],[iter3])
GO
"""
    assert len(_split(seed)) == 3


def test_value_table_selection():
    seed = """INSERT INTO mytable SELECT * FROM ( VALUES (133713371337), (133713371338) ) t(id); GO"""
    actual = _split(seed)
    expected = """INSERT INTO mytable SELECT * FROM ( VALUES ( 133713371337 ) , ( 133713371338 ) ) t ( id ) ;"""
    assert len(actual) == 1
    assert_strings_equal_disregarding_whitespace(actual[0], expected)


def test_select_from():
    seed = """
SELECT CONCAT('0', Nr) nr
       , nr19
       , MIN(VALID_FROM) min_valid_from
       , ROW_NUMBER() OVER(PARTITION BY Nr, MIN(VALID_FROM) ORDER BY COUNT(distinct VALID_FROM) DESC, nr19) iter
  FROM otherdb.dbo.history WITH(NOLOCK)
  CROSS APPLY ( VALUES (
    CASE LEFT(NR_PROZEDUR, 1) WHEN '5' THEN LEFT(NR_PROZEDUR, 6) + SUBSTRING(NR_PROZEDUR, 9, 13)
    ELSE LEFT(NR_PROZEDUR, 19) END
  )) nr(nr19)
  GROUP BY Nr, nr19
"""
    assert len(_split(seed)) == 1


def test_create_index_include():
    seed = """IF NOT
    (
        SELECT MIN(ENGBE/ENA)         -- Awesome comment.
        FROM dbo.output WITH(NOLOCK)
        WHERE LEFT(code, 4) = 'code20' -- case-specific codes start with code20.
    )
    >
    (
        SELECT unspecified_engbe_placeholder - epsilon
        FROM dbo.vars WITH(NOLOCK)
    )
  AND
    (
        SELECT MAX(ENGBE/ENA)         -- More comment.
        FROM dbo.output WITH(NOLOCK)
        WHERE LEFT(code, 4) = 'code20' -- case-specific codes start with code20.
    )
    <
    (
        SELECT unspecified_engbe_placeholder + epsilon
        FROM dbo.vars WITH(NOLOCK)
    )
BEGIN
    RAISERROR('Inconsistent placeholder ', 18, 16)
    RETURN
END
"""
    assert len(_split(seed)) == 1


def test_redundant_go():
    seed = """
        GO
        DROP TABLE IF EXISTS dbo.mytable
        GO
    """
    assert len(_split(seed)) == 1


def test_scalar_functions():
    seed = """ SELECT SQRT(SQUARE(SIGN(LOG((EXP(3/0.4))))))
    FROM table2;
    """
    assert len(_split(seed)) == 1


def test_synonym_drop():
    seed = """
    DROP SYNONYM IF EXISTS hax
    DROP SYNONYM IF EXISTS dirty.hax"""
    assert len(_split(seed)) == 2


def test_synonym_create():
    seed = """
    CREATE SYNONYM hax FOR dirty.hax;
    CREATE SYNONYM dirty.hax FOR leet.hax;
    """
    assert len(_split(seed)) == 2


def test_foreign_key():
    seed = """
    CREATE TABLE dbo.mytable (
    id BIGINT NOT NULL
    otherid BIGINT NOT NULL
    , PRIMARY KEY (id, otherid)
    , FOREIGN KEY (otherid) REFERENCES dbo.othertable(id)
);
"""
    assert len(_split(seed)) == 1


def test_parser_error():
    with pytest.raises(ValueError):
        _split("USEE master")


def test_lexer_error():
    with pytest.raises(ValueError):
        _split("USE ?")


def test_go_as_alias():
    with pytest.raises(ValueError):
        _split("SELECT * FROM (VALUES(1)) AS GO(go)")


def test_xml_value():
    seed = """
    SELECT my_id
         , STUFF((
                     SELECT ', ' + other_table.display
                     FROM other_table
                     WHERE other_table.my_id = my_table.my_id
                     ORDER BY other_table.iter
                     FOR XML PATH(''),TYPE
                 ).value('(./text())[1]', 'VARCHAR(MAX)'), 1, 2,
                 '') eps
    -- type conversion via .value, remove 2 char at position 1 via STUFF to strip leading delimiter
    FROM my_table
    GROUP BY my_table.my_id
    """
    _split(seed)


def test_xml_methods():
    seed = """
    DECLARE @myDoc XML
    SELECT @myDoc.query('XQuery')
    SELECT @myDoc.value('XQuery', 'int')
    SELECT @myDoc.exist('XQuery')
    SELECT @myDoc.nodes('XQuery') AS T(C)
    """
    assert len(_split(seed)) == 5


def test_non_xml_method():
    seed = """
    DECLARE @myDoc XML
    SELECT @myDoc.fakemethod('param')
    """
    with pytest.raises(ValueError):
        _split(seed)


def test_linked_server():
    seed = """
    SELECT *
    INTO linked_server.db.schm.table1
    FROM linked_server.db.schm.table2
    """
    assert len(_split(seed)) == 1


def test_rowcount_no_isolation():
    seed = """
    USE new_db
    PRINT('Using new_db')
    GO
    SELECT * FROM my_table
    PRINT(@@ROWCOUNT)
    """
    assert len(_split(seed)) == 4
    assert len(_split(seed, isolate_top_level_statements=False)) == 2


def test_type_as_table():
    seed = """
    CREATE TYPE x AS TABLE (
        id INT
    )
    """
    assert len(_split(seed)) == 1



import pytest

from pytsql.tsql import _split


# Note, Kevin, 21/12/03
# As of now, there seems to be a problem with this query.
# This is not by design. We would like to succeed this test.
@pytest.mark.xfail
def test_smoke_pivot(engine):
    seed = """
CREATE VIEW my_view AS (
    SELECT myid
           , points_derivative
           , feature1
           , feature2
    FROM (
        SELECT myid
             , points
             , criterion
             , SUM(points) OVER(PARTITION BY myid) points_derivative
        FROM table1
    ) SourceTable
    PIVOT (
        AVG(points)
        FOR criterion IN (
            feature1
          , feature2
        )
    ) PivotTable
)
GO
    """
    _split(seed)



"""`Pytsql` allows to run mssql sripts, typically run via GUIs, via CLI."""

import pkg_resources

from .tsql import execute, executes

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except Exception:
    __version__ = "unknown"

__all__ = ["execute", "executes"]



# Configuration file for the Sphinx documentation builder.

# -- Project information
project = "pytsql"
copyright = "(C) 2020-2021 QuantCo Inc."
author = "QuantCo Inc."

release = "1.0"
version = "1.0.0"

extensions = ["numpydoc", "sphinxcontrib.apidoc"]

apidoc_module_dir = "../../src/pytsql"
apidoc_output_dir = "api"
apidoc_separate_modules = True
apidoc_extra_args = ["--implicit-namespaces"]
apidoc_excluded_paths = ["grammar"]

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



from typing import Optional

import pytest
import sqlalchemy as sa
from sqlalchemy.engine import Engine
from sqlalchemy.exc import ProgrammingError

from pytsql.tsql import executes


def test_multiple_statements_middle_should_break_due_to_existing_table(engine):
    seed = """
USE [tempdb]

DROP TABLE IF EXISTS [test_table];

CREATE TABLE [test_table] (
col VARCHAR(3)
)
CREATE TABLE [test_table] (
col VARCHAR(3)
)
INSERT INTO [test_table]
SELECT 'ok';
GO

DROP TABLE IF EXISTS [test_table];
GO
    """
    with pytest.raises(
        ProgrammingError,
        match="There is already an object named 'test_table' in the database.",
    ):
        executes(seed, engine, None)


def test_failing_insertion(engine):
    seed = """
USE master;
DROP TABLE IF EXISTS test_table;
CREATE TABLE test_table (
    col_test VARCHAR(3)
);
INSERT INTO test_table
SELECT 'string causing a silent failure'
GO
    """
    with pytest.raises(
        ProgrammingError, match="String or binary data would be truncated"
    ):
        executes(seed, engine, None)


def test_multiple_statements_with_spurius_go(engine):
    statement = """
USE tempdb
GO
DROP TABLE IF EXISTS test_schema.test_table
DROP schema IF EXISTS test_schema
CREATE schema test_schema
GO
CREATE TABLE test_schema.test_table (
 id BIGINT NOT NULL PRIMARY KEY
)
INSERT INTO test_schema.test_table WITH(TABLOCKX)
SELECT *
FROM (
 VALUES (123)
) a(id)
GO
    """
    executes(statement, engine, None)

    table = get_table(engine, "test_table", "test_schema")
    assert len(table.primary_key.columns) == 1
    assert "id" in table.primary_key.columns
    assert str(table.columns["id"].type) == "BIGINT"

    with engine.connect() as conn:
        res = conn.execute("SELECT * FROM test_schema.test_table")
        rows = res.all()
        assert len(rows) == 1
        assert len(rows[0]) == 1
        assert rows[0][0] == 123


def test_primary_key_creation(engine):
    statement = """
    DROP DATABASE IF EXISTS primary_key_creation
    CREATE DATABASE primary_key_creation
    USE primary_key_creation
GO

/****** Object:  Table [dbo].[table1]    Script Date: 2/23/2021 2:48:02 PM ******/
SET ANSI_NULLS ON
GO

SET QUOTED_IDENTIFIER ON
GO

CREATE TABLE [dbo].[table2](
    [nr] [varchar](17) NOT NULL,
    [valid_from] [date] NOT NULL,
    [load_date] [date] NOT NULL,
PRIMARY KEY NONCLUSTERED
(
    [nr] ASC,
    [valid_from] ASC
)WITH (PAD_INDEX = OFF, STATISTICS_NORECOMPUTE = OFF, IGNORE_DUP_KEY = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) ON [PRIMARY]
) ON [PRIMARY]
GO"""
    executes(statement, engine, None)

    table = get_table(engine, "table2", "primary_key_creation.dbo")
    assert len(table.primary_key.columns) == 2
    assert "nr" in table.primary_key.columns
    assert "valid_from" in table.primary_key.columns
    assert "VARCHAR(17)" in str(table.columns["nr"].type)
    assert str(table.columns["valid_from"].type) == "DATE"
    assert str(table.columns["load_date"].type) == "DATE"


def test_multiple_uses(engine):
    seed = """
    DROP DATABASE IF EXISTS test_database
    DROP TABLE IF EXISTS master.dbo.table2
    DROP TABLE IF EXISTS master.dbo.table3
    CREATE DATABASE test_database
    USE test_database
    CREATE TABLE dbo.table1 (
        col_test VARCHAR(3)
    ) -- Created into test_database

    USE master
    CREATE TABLE dbo.table2 (
        col_test VARCHAR(3)
    ) -- Created into master

    CREATE TABLE dbo.table3 (
        col_test VARCHAR(3)
    ) -- Created into master because that's the last USE stmt

    CREATE TABLE test_database.dbo.table4(
        col_test VARCHAR(3)
    ) -- Created into test_database because the db is qualified
    GO
    """
    executes(seed, engine, None)

    master_names = sa.inspect(engine).get_table_names(schema="dbo")
    test_database_names = sa.inspect(engine).get_table_names(schema="test_database.dbo")

    assert "table2" in master_names
    assert "table3" in master_names
    assert "table1" in test_database_names
    assert "table4" in test_database_names


def get_table(
    engine: Engine, table_name: str, schema: Optional[str] = None
) -> sa.Table:
    return sa.Table(table_name, sa.MetaData(), autoload_with=engine, schema=schema)



import logging

from pytsql.tsql import _split

from .helpers import assert_strings_equal_disregarding_whitespace

LOGGER = logging.getLogger(__name__)


def test_top_level_print():
    seed = """
    SELECT TOP 10 * FROM my_table;
    PRINT 'text';
    SELECT TOP 10 * FROM my_table2;
    """

    splits = _split(seed)

    assert splits == [
        "SELECT TOP 10 * FROM my_table ;",
        "INSERT INTO #pytsql_prints VALUES ( LEFT ( 'text' , 2000 ) ) ;",
        "SELECT TOP 10 * FROM my_table2 ;",
    ]


def test_prints_in_control_flow():
    seed = """
    DECLARE @A INT = 4
    IF 1 = 1
    BEGIN
        PRINT(2 + 2)
        PRINT @A
    END
    """.strip()

    expected = """
    DECLARE @A INT = 4
    IF 1 = 1
        BEGIN
            INSERT INTO #pytsql_prints VALUES ( LEFT ( ( 2 + 2 ) , 2000 ) ) ;
            INSERT INTO #pytsql_prints VALUES ( LEFT ( @A , 2000 ) ) ;
        END
    """.strip()

    splits = _split(seed)

    assert len(splits) == 2
    assert_strings_equal_disregarding_whitespace(splits[1], expected)



import pytest

from pytsql.tsql import _parametrize


def test_replace_comment():
    seed = """
/* <replace> */
select a
    , b
    , c
from
    x
/* </replace> select {alpha},b,{charlie} from x */
    """  # noqa: W291
    expected = """select first,b,second from x"""
    assert _parametrize(seed, {"alpha": "first", "charlie": "second"}) == expected


def test_replace_comment_second():
    seed = """
USE master;
/* <replace> */
select * from [master].[dbo].[table];
/* </replace> select * from [master].[dbo].[{tableName}]; */
    """
    expected = """USE master;
select * from [master].[dbo].[new_table];"""
    assert _parametrize(seed, {"tableName": "new_table"}) == expected


def test_replace_comment_int():
    seed = """
USE master;
/* <replace> */
select top 42 * from [master].[dbo].[table];
/* </replace> select top {n_rows} * from [master].[dbo].[table]; */
    """
    expected = """USE master;
select top 1337 * from [master].[dbo].[table];"""
    assert _parametrize(seed, {"n_rows": 1337}) == expected


def test_replace_comment_class():
    class MyClass:
        def __init__(self, my_table: str):
            self.my_table = my_table

        def __str__(self):
            return "master.dbo." + self.my_table

    seed = """
USE master;
/* <replace> */
select * from [master].[dbo].[table];
/* </replace> select * from {table_qualifier}; */
    """
    expected = """USE master;
select * from master.dbo.table;"""
    assert _parametrize(seed, {"table_qualifier": MyClass("table")}) == expected


def test_none_replacement():
    seed = """
USE master;
/* <replace> */
select * from [master].[dbo].[table];
/* </replace> select * from [master].[dbo].[{tableName}]; */
    """
    with pytest.raises(ValueError):
        _parametrize(seed, {"tableName": None})


def test_double_replacement():
    seed = """
USE master;
SELECT * 
FROM dbo./* <replace> */table/* </replace>{tableName} */ t1
JOIN dbo./* <replace> */table/* </replace>{tableName} */ t2
ON t1.id = t2.id;
"""  # noqa: W291
    expected = """USE master;
SELECT *
FROM dbo.new_table t1
JOIN dbo.new_table t2
ON t1.id = t2.id;"""  # noqa: W291
    assert _parametrize(seed, {"tableName": "new_table"}) == expected


def test_multiline_replacement():
    seed = """
USE master;
/* <replace> */
select * from [master].[dbo].[table];
/* </replace> select * from [master].[dbo].[{tableName}]; */
    """
    expected = """USE master;
select * from [master].[dbo].[new_table];"""
    assert _parametrize(seed, {"tableName": "new_table"}) == expected


def test_multi_multiline_replacements():
    seed = """
USE master;
/* <replace> */
select * from [master].[dbo].[table];
/* </replace> select * from [master].[dbo].[{tableName}]; */
USE newTable;
/* <replace> */
select * from [master].[dbo].[table];
/* </replace> select * from [master].[dbo].[{otherTable}]; */
    """  # noqa: W291
    expected = """USE master;
select * from [master].[dbo].[new_table];
USE newTable;
select * from [master].[dbo].[second_table];"""  # noqa: W291
    assert (
        _parametrize(seed, {"tableName": "new_table", "otherTable": "second_table"})
        == expected
    )


def test_regular_comment_stays():
    seed = """
USE master; /* this is a regular comment */
select * from [master].[dbo].[table];
    """
    expected = """USE master; /* this is a regular comment */
select * from [master].[dbo].[table];"""
    assert _parametrize(seed, {"branch": "new_master"}) == expected


def test_unknown_parameter_exception():
    seed = """
/* <replace> */
USE master;
/* </replace> USE {randomTable}; */
select * from [master].[dbo].[table];
    """
    with pytest.raises(KeyError):
        _parametrize(seed, {"key": "value"})


def test_same_line_replacement():
    seed = """
select a,b,/* <replace> */c/* </replace>{otherC} */ from x
    """
    expected = """select a,b,newC from x"""
    assert _parametrize(seed, {"otherC": "newC"}) == expected


def test_multi_same_line_replacements():
    seed = """
SELECT /* <replace> */a/* </replace>{newA} */, /* <replace> */b/* </replace>{newB} */
FROM table;
    """
    expected = """SELECT newA, newB
FROM table;"""
    assert _parametrize(seed, {"newA": "newA", "newB": "newB"}) == expected


def test_replace_in_string_literal():
    seed = """SELECT '/* <replace> */a/* </replace>{newA} */'"""
    assert _parametrize(seed, {"newA": "newA"}) == "SELECT 'newA'"


def test_custom_replace_keywords():
    seed = """
    SELECT a,/* <replace> */b/* </replace>{otherB} */,/* *repl_start* */c/* *repl_end*{otherC} */ FROM x
        """
    expected = """SELECT a,/* <replace> */b/* </replace>{otherB} */,newC FROM x"""
    assert (
        _parametrize(
            seed,
            {"otherB": "newB", "otherC": "newC"},
            start="*repl_start*",
            end="*repl_end*",
        )
        == expected
    )



import glob
import platform
from time import time

import setuptools
from setuptools_scm.version import ScmVersion


def get_dev_timestamp(version: ScmVersion) -> str:
    """Return a new distribution version string.

    Returns the version found in the git tag if currently checked out commit has a tag.
    Otherwise, returns the version found in the most recent git tag, appended with
    `dev` and the current timestamp in seconds.
    """

    if version.exact:
        return version.format_with("{tag}")
    return version.format_with(f"{{tag}}.dev{int(time())}")


def get_platform() -> str:
    """Return the platform on which this code is run.

    Returns the platform name stripped of unnecessary information if the platform is known.
    Otherwise, returns the full platform name.
    """

    known_platforms = {"windows", "linux", "darwin", "cygwin"}

    target = platform.system().lower()
    for known in known_platforms:
        if target.startswith(known):
            return known

    return target


def create_extension() -> setuptools.Extension:
    """Create the ANTLR C++ extension to be passed to `setuptools.setup`"""

    extra_compile_args = {
        "windows": ["/DANTLR4CPP_STATIC", "/Zc:__cplusplus"],
        "linux": ["-std=c++11"],
        "darwin": ["-std=c++11"],
        "cygwin": ["-std=c++11"],
    }

    sources = glob.glob("src/pytsql/grammar/cpp_src/**/*.cpp", recursive=True)
    depends = glob.glob("src/pytsql/grammar/cpp_src/**/*.h", recursive=True)

    return setuptools.Extension(
        name="pytsql.grammar.sa_tsql_cpp_parser",
        include_dirs=["src/pytsql/grammar/cpp_src/antlr4-cpp-runtime"],
        sources=sources,
        depends=depends,
        extra_compile_args=extra_compile_args.get(get_platform(), []),
    )


def run_setup():
    setuptools.setup(
        ext_modules=[create_extension()],
        use_scm_version={
            "version_scheme": get_dev_timestamp,
            "local_scheme": "dirty-tag",
        },
    )


run_setup()



import os
import urllib.parse

import pytest
import sqlalchemy as sa

TEST_DB_NAME = "tempdb"


def get_engine(backend) -> sa.engine.Engine:
    address = os.environ.get("DB_ADDR", "localhost")
    # auth = "sa:QuantCo%40MSSQL@" if backend != "mssql-windows-ci" else ""
    connection_string = (
        f"mssql+pyodbc://sa:pytsql-Test-123@{address}:1433/{TEST_DB_NAME}"
    )
    if backend == "mssql-freetds":
        connection_string += "?driver=libtdsodbc.so&tds_version=7.4"
    else:
        msodbc_driver_name = urllib.parse.quote_plus("ODBC Driver 17 for SQL Server")
        connection_string += f"?driver={msodbc_driver_name}"
    return sa.create_engine(
        connection_string, connect_args={"autocommit": True, "timeout": 10}
    )


@pytest.fixture(scope="function")
def engine(backend):
    yield get_engine(backend)


def pytest_addoption(parser):
    parser.addoption(
        "--backend",
        choices=("mssql", "mssql-freetds", "mssql-windows-ci"),
        help="which database backend to use to run the integration tests",
    )


def pytest_generate_tests(metafunc):
    if "backend" in metafunc.fixturenames:
        metafunc.parametrize(
            "backend", [metafunc.config.getoption("backend")], scope="module"
        )
